// file      : xsd/cxx/tree/stream-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/stream-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c,
              String const& name_ = "",
              String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& l)
        {
          Boolean named (l.named ());

          if (named)
          {
            name = ename (l);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          os << esymbol
             << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c,
               String const& name_ = "",
               String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& u)
        {
          Boolean named (u.named ());

          if (named)
          {
            name = ename (u);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          os << esymbol
             << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      struct Enumeration: Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     String const& name_ = "",
                     String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& e)
        {
          Boolean named (e.named ());

          if (named)
          {
            name = ename (e);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          os << esymbol
             << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >&," << endl
             << type_name << "::_xsd_" <<  name << ");"
             << endl;

          os << esymbol
             << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >&," << endl
             << (named ? "class " : "") << type_name << ");"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            // I need to build scope information with the same names
            // as in the header so that I get member escaping right.
            // Actual names will be from the public interface.
            //

            String const& name (ename (m));
            String type_name (scope
                              + L"::" + name
                              + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (L"_xsd_" + name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }

      private:
        String scope;
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c,
                 String const& name_ = "",
                 String const& type_name_ = L"")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean named (c.named ());

          if (named)
          {
            name = ename (c);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // Go after anonymous types first.
          //
          {
            enter_scope (name);

            Traversal::Names names_member;
            Member<Complex> member (*this, type_name);

            names_member >> member;

            names (c, names_member);

            leave_scope ();
          }

          os << esymbol
             << "::std::basic_ostream< " << char_type << " >&" << endl
             << "operator<< (::std::basic_ostream< " << char_type <<
            " >&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element, GlobalElementBase
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String const& name (ename (m));
            String type_name (name + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };
    }

    Void
    generate_stream_header (Context& ctx, SemanticGraph::Schema& tu)
    {
      String c (ctx.char_type);

      ctx.os << "#include <iosfwd> // std::basic_ostream&" << endl
             << endl;


      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (tu);
    }
  }
}
