// file      : xsd/cxx/tree/tree-inline.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-inline.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
            name = ename (l);

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " ()"
             << "{"
             << "}";

          // copy c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " (class " << name << " const& o," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (o, f, c)," << endl
             << "::xsd::cxx::tree::sequence< " <<
            type_name (l.argumented ().type ()) << " > (o, f, c)"
             << "{"
             << "}";
        }

      private:
        String
        type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
            name = ename (u);

          String base (L"::xsd::cxx::tree::string< " + char_type + L" >");

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor (string const&)
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " (" << string_type << " const& s)" << endl
             << ": " << base << " (s)"
             << "{"
             << "}";

          // copy c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " (class " << name << " const& o," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (o, f, c)"
             << "{"
             << "}";
        }

      private:
        String name;
      };


      struct Enumeration : Traversal::Enumeration,
                           protected virtual Context
      {
        Enumeration (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = ename (e);

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "// " << endl
             << endl;

          // c-tor (_xsd_name)
          //
          os << inl
             << scoped_name << "::" <<  endl
             << name << " (_xsd_" << name << " v)" << endl
             << ": _xsd_" << name << "_ (v)" << endl
             << "{"
             << "}";


          // copy c-tor
          //
          os << inl
             << scoped_name << "::" <<  endl
             << name << " (class " << name << " const& v," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (v, f, c)," << endl
             << "_xsd_" << name << "_ (v._xsd_" << name << "_)"
             << "{"
             << "}";

          // operator= (_xsd_name)
          //
          os << inl
             << "class " << scoped_name << "& " << scoped_name << "::" << endl
             << "operator= (_xsd_" << name << " v)"
             << "{"
             << "_xsd_" << name << "_ = v;"
             << "return *this;"
             << "}";

          os << endl;
        }

      private:
        String name;
      };

      struct MemberFunction: Traversal::Member, protected virtual Context
      {
        MemberFunction (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          if (max (m) != 1)
          {
            // sequence
            //
            String container (iscope_name (name, "container"));
            String q_container (scope + L"::" + container);

            // container const&
            // name () const;
            //
            os << inl
               << q_container << " const& " << scope << "::" << endl
               << name << " () const"
               << "{"
               << "return this->_xsd_" << name << "_ ();"
               << "}";

            // container&
            // name ();
            //
            os << inl
               << q_container << "& " << scope << "::" << endl
               << name << " ()"
               << "{"
               << "return this->_xsd_" << name << "_ ();"
               << "}";

            // void
            // name (container const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (" << container << " const& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_ (" << name << ");"
               << "}";
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            String type (iscope_name (name, "type"));
            String container (iscope_name (name, "container"));
            String q_container (scope + L"::" + container);

            // container const&
            // name () const;
            //
            os << inl
               << q_container << " const& " << scope << "::" << endl
               << name << " () const"
               << "{"
               << "return this->_xsd_" << name << "_ ();"
               << "}";

            // container&
            // name ();
            //
            os << inl
               << q_container << "& " << scope << "::" << endl
               << name << " ()"
               << "{"
               << "return this->_xsd_" << name << "_ ();"
               << "}";

            // void
            // name (type const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (" << type << " const& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_ (" << name << ");"
               << "}";

            // void
            // name (container const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (" << container << " const& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_ (" << name << ");"
               << "}";

            // void
            // name (auto_ptr<type>);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (::std::auto_ptr< " << type << " > " <<
              name << ")"
               << "{"
               << "this->_xsd_" << name << "_ (" << name << ");"
               << "}";
          }
          else
          {
            // one
            //
            String type (iscope_name (name, "type"));
            String q_type (scope + L"::" + type);

            // type const&
            // name () const;
            //
            os << inl
               << q_type << " const& " << scope << "::" << endl
               << name << " () const"
               << "{"
               << "return this->_xsd_" << name << "_ ();"
               << "}";

            // type&
            // name ();
            //
            os << inl
               << q_type << "& " << scope << "::" << endl
               << name << " ()"
               << "{"
               << "return this->_xsd_" << name << "_ ();"
               << "}";

            // void
            // name (type const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (" << type << " const& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_ (" << name << ");"
               << "}";

            // void
            // name (auto_ptr<type>);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (::std::auto_ptr< " << type << " > " <<
              name << ")"
               << "{"
               << "this->_xsd_" << name << "_ (" << name << ");"
               << "}";
          }
        }
      };

      template <typename Complex>
      struct Member : Traversal::Member,
                      protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              member_function_ (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          SemanticGraph::Type& t (m.type ());

          // Go after anonymous type first.
          //
          if (!t.named () && !t.context ().count ("seen"))
          {
            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (L"_xsd_" + name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          // default_value
          //
          if (m.default_ ())
          {
            Boolean simple (true);

            if (m.is_a<SemanticGraph::Element> ())
            {
              IsSimpleType test (simple);
              test.dispatch (t);
            }

            if (simple)
            {
              String type (iscope_name (name, "type"));
              String q_type (scope + L"::" + type);

              os << inl
                 << q_type << " const& " << scope << "::" <<
                name << "::" << endl
                 << unclash (name, "default_value") << " ()"
                 << "{"
                 << "return " << unclash (name, "default_value") << "_;"
                 << "}";
            }
          }

          member_function_.traverse (m);
        }

      private:
        MemberFunction member_function_;
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              member_ (c)
        {
          names_ >> member_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = ename (c);

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "// " << endl
             << endl;


          // Go after anonymous types and generate accessors and modifiers
          // first.
          //
          {
            enter_scope (name);
            names (c, names_);
            leave_scope ();
          }

          os << endl;
        }

      private:
        String name;

        Traversal::Names names_;
        Member<Complex> member_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element, GlobalElementBase
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String const& name (ename (m));

            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };

      struct Includes : Traversal::Imports,
                        Traversal::Includes,
                        protected virtual Context
      {
        Includes (Context& c, String const& expr)
            : Context (c), expr_ (expr)
        {
        }

        virtual Void
        traverse (SemanticGraph::Imports& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Includes& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Path const& path)
        {
          using BackendElements::Regex::perl_s;

          // Note the use of the portable representation of the path.
          //
          String name (perl_s (String (path.string ()), expr_));

          Char op (include_with_brackets ? '<' : '"');
          Char cl (include_with_brackets ? '>' : '"');

          os << "#include " << op << include_prefix << name << cl << endl
             << endl;
        }

      private:
        String expr_;
      };
    }

    Void
    generate_tree_inline (Context& ctx,
                          SemanticGraph::Schema& tu,
                          String const& expr,
                          UnsignedLong first,
                          UnsignedLong last)
    {
      // Generate includes.
      //
      if (ctx.options.value<CLI::generate_inline> ())
      {
        Traversal::Schema schema;
        Includes includes (ctx, expr);

        schema >> includes;

        schema.dispatch (tu);
      }


      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx, false, first, last);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (tu);
    }
  }
}
